import torch
import torch.nn as nn
import torch.optim as optim

import torch.nn.functional as F
import torchaudio
import tqdm

from torch.utils.data import random_split
from torch.utils.data import DataLoader

from data import AudioDataset
from generator import SISUGenerator
from discriminator import SISUDiscriminator

# Check for CUDA availability
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(f"Using device: {device}")

# Initialize dataset and dataloader
dataset_dir = './dataset/good'
dataset = AudioDataset(dataset_dir, target_duration=2.0)

dataset_size = len(dataset)
train_size = int(dataset_size * .9)
val_size = int(dataset_size-train_size)

train_dataset, val_dataset = random_split(dataset, [train_size, val_size])

train_data_loader = DataLoader(train_dataset, batch_size=8, shuffle=True)
val_data_loader = DataLoader(val_dataset, batch_size=8, shuffle=True)

# Initialize models and move them to device
generator = SISUGenerator()
discriminator = SISUDiscriminator()

generator = generator.to(device)
discriminator = discriminator.to(device)

# Loss
criterion_g = nn.L1Loss()
criterion_d = nn.BCEWithLogitsLoss()

# Optimizers
optimizer_g = optim.Adam(generator.parameters(), lr=0.0001, betas=(0.5, 0.999))
optimizer_d = optim.Adam(discriminator.parameters(), lr=0.0001, betas=(0.5, 0.999))

# Scheduler
scheduler_g = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_g, mode='min', factor=0.5, patience=5)
scheduler_d = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_d, mode='min', factor=0.5, patience=5)

def snr(y_true, y_pred):
    noise = y_true - y_pred
    signal_power = torch.mean(y_true ** 2)
    noise_power = torch.mean(noise ** 2)
    snr_db = 10 * torch.log10(signal_power / noise_power)
    return snr_db

def discriminator_train(discriminator, optimizer, criterion, generator, real_labels, fake_labels, high_quality, low_quality):
    optimizer.zero_grad()

    discriminator_decision_from_real = discriminator(high_quality)
    d_loss_real = criterion(discriminator_decision_from_real, real_labels)

    generator_output = generator(low_quality)
    discriminator_decision_from_fake = discriminator(generator_output.detach())
    d_loss_fake = criterion(discriminator_decision_from_fake, fake_labels)

    d_loss = (d_loss_real + d_loss_fake) / 2.0

    d_loss.backward()
    nn.utils.clip_grad_norm_(discriminator.parameters(), max_norm=1.0) #Gradient Clipping
    optimizer.step()
    # print(f"Discriminator Loss: {d_loss.item():.4f}, Mean Real Logit: {discriminator_decision_from_real.mean().item():.2f}, Mean Fake Logit: {discriminator_decision_from_fake.mean().item():.2f}")

def start_training():

    # Training loop
    # discriminator_epochs = 1000
    generator_epochs = 500
    for generator_epoch in range(generator_epochs):
        low_quality_audio = torch.empty((1))
        high_quality_audio = torch.empty((1))
        ai_enhanced_audio = torch.empty((1))

        # Training
        for low_quality, high_quality in tqdm.tqdm(train_data_loader, desc=f"Epoch {generator_epoch+1}/{generator_epochs}"):
            high_quality = high_quality.to(device)
            low_quality = low_quality.to(device)

            batch_size = high_quality.size(0)
            real_labels = torch.ones(batch_size, 1).to(device)
            fake_labels = torch.zeros(batch_size, 1).to(device)

            # Train Discriminator
            discriminator.train()

            for _ in range(3):
                discriminator_train(discriminator, optimizer_d, criterion_d, generator, real_labels, fake_labels, high_quality, low_quality)

            # Train Generator
            generator.train()
            optimizer_g.zero_grad()

            # Generator loss: how well fake data fools the discriminator
            generator_output = generator(low_quality)
            discriminator_decision = discriminator(generator_output)  # No detach here
            g_loss = criterion_g(discriminator_decision, real_labels)  # Train generator to produce real-like outputs

            g_loss.backward()
            optimizer_g.step()

            low_quality_audio = low_quality
            high_quality_audio = high_quality
            ai_enhanced_audio = generator_output

        metric = snr(high_quality_audio, ai_enhanced_audio)
        print(f"Generator metric {metric}!")
        scheduler_g.step(metric)

        if generator_epoch % 10 == 0:
            print(f"Saved epoch {generator_epoch}!")
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-crap.wav", low_quality_audio[0].cpu(), 44100)
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-ai.wav", ai_enhanced_audio[0].cpu(), 44100)
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-orig.wav", high_quality_audio[0].cpu(), 44100)

        if generator_epoch % 50 == 0:
            torch.save(discriminator.state_dict(), "discriminator.pt")
            torch.save(generator.state_dict(), "generator.pt")

    torch.save(discriminator.state_dict(), "discriminator.pt")
    torch.save(generator.state_dict(), "generator.pt")
    print("Training complete!")

start_training()
